from rich.traceback import Traceback as RichTraceback, Stack, PathHighlighter, WINDOWS, _SyntaxError
from rich.highlighter   import ReprHighlighter
from rich.padding   import Padding
from rich.console   import RenderResult, group as render_group
from rich.text      import Text
from rich.syntax    import Syntax


# Patch Traceback to print Syntax from cells rather than from_path
class Traceback(RichTraceback):
    def __init__(self, nb, debug = False, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.nb = nb

        if not debug:
            # skip past exec_eval
            for i,stack in enumerate(self.trace.stacks):
                eval_location = next((j for j,x in enumerate(stack.frames) if x.name == 'exec_eval'), None)
                if eval_location is not None:
                    self.trace.stacks[i].frames = stack.frames[eval_location+1:]

    @staticmethod
    def filename_cell(fn):
        fn_split = fn.split(':')
        if len(fn_split) == 1:
            return fn, -1, -1
        filename = fn_split[0]
        cell_id = int(fn_split[1])
        cell_id_display = int(fn_split[2])
        return filename, cell_id, cell_id_display

    @render_group()
    def _render_stack(self, stack: Stack) -> RenderResult:
        path_highlighter = PathHighlighter()
        theme = self.theme or ("fruity" if WINDOWS else "monokai")
        for frame in stack.frames:
            filename, cell_id, cell_id_display = self.filename_cell(frame.filename)
            if cell_id != -1:
                text = Text.assemble(
                    (" File ", "traceback.text"),
                    (f'"{filename}"', "traceback.filename"),
                    (", cell ", "traceback.text"),
                    (f"{cell_id_display}", "traceback.lineno"),
                    (", line ", "traceback.text"),
                    (str(frame.lineno), "traceback.lineno"),
                    (", in ", "traceback.text"),
                    (frame.name, "traceback.name"),
                )
            else:
                text = Text.assemble(
                    (" File ", "traceback.text"),
                    (f'"{filename}"', "traceback.filename"),
                    (", line ", "traceback.text"),
                    (str(frame.lineno), "traceback.lineno"),
                    (", in ", "traceback.text"),
                    (frame.name, "traceback.name"),
                )
            yield path_highlighter(text)
            if frame.filename.startswith("<"):
                continue
            try:
                if cell_id != -1:
                    syntax = Syntax(
                        self.nb.cells[cell_id].code(),
                        'python',
                        theme=theme,
                        line_numbers=True,
                        line_range=(
                            frame.lineno - self.extra_lines,
                            frame.lineno + self.extra_lines,
                        ),
                        highlight_lines={frame.lineno},
                        word_wrap=self.word_wrap,
                    )
                else:
                    syntax = Syntax.from_path(
                        filename,
                        theme=theme,
                        line_numbers=True,
                        line_range=(
                            frame.lineno - self.extra_lines,
                            frame.lineno + self.extra_lines,
                        ),
                        highlight_lines={frame.lineno},
                        word_wrap=self.word_wrap,
                    )
            except Exception:
                raise
            else:
                yield Padding.indent(syntax, 2)



from matplotlib._pylab_helpers import Gcf
from matplotlib.backend_bases import (FigureCanvasBase, FigureManagerBase)
from matplotlib.backends.backend_agg import FigureCanvasAgg

import io

figures = None

class SaturnFigureManager(FigureManagerBase):
    def show(self):
        with io.BytesIO() as buf:
            self.canvas.figure.savefig(buf, format='png')
            figures.append_png(buf.getvalue())

class SaturnCanvas(FigureCanvasAgg):
    manager_class = SaturnFigureManager

FigureCanvas  = SaturnCanvas
FigureManager = SaturnFigureManager

def show(*, block=None):
    for manager in Gcf.get_all_fig_managers():
        manager.show()
    Gcf.destroy_all()



from ptpython.repl import PythonRepl

class PythonReplWithExecute(PythonRepl):
    def __init__(self, execute, *args, debug=False, **kw):
        self.execute__ = execute
        super().__init__(*args, **kw)

        self.show_meta_enter_message = False
        self.title = "Saturn "
        self.use_code_colorscheme('vim')

        self.complete_while_typing = False
        self.enable_history_search = True
        self.enable_auto_suggest = True

        self.confirm_exit = False

        # Swap enter and meta+enter #
        @self.add_key_binding('escape', 'enter')
        def _(event):
            event.current_buffer.insert_text('\n')

        @self.add_key_binding('enter', filter = self.show_exit_confirmation)
        def _(event):
            event.current_buffer.validate_and_handle()

        @self.add_key_binding('c-k')
        def _(event):
            event.app.current_buffer.insert_text("#chk>")
            event.current_buffer.validate_and_handle()

        # Unless we are in debug mode, silence ptpython's exception handling, in favor of our own
        if not debug:
            self._handle_exception = lambda e: None

    def eval(self, line: str) -> None:
        self.execute__(line)



import  os, sys
import  argh
from    rich.console import Console
from    rich.rule    import Rule
from    rich.panel   import Panel

import  io
from    atomicwrites import atomic_write
from    more_itertools import peekable

from    contextlib import nullcontext

from    .           import cells as c, notebook
from    .repl       import PythonReplWithExecute
from    .image      import show_png

root = True
using_mpi = False

skip_repl = False

# workaround for a bug in OpenMPI (or anything else that screws up the terminal size);
# see https://github.com/willmcgugan/rich/issues/127
import  shutil
width = None if shutil.get_terminal_size().columns != 0 else 80

from    .theme      import theme
console = Console(width = width, theme = theme)

argv = []

def info(*args, block = False, **kw):
    if root:
        if not block:
            console.print(Rule(*args, **kw))
        else:
            console.print(*args, **kw)

def error(*args, **kw):
    if root:
        console.print(Panel(*args, **kw, style='error'))

def warn(*args, **kw):
    if root:
        console.print(Panel(*args, **kw, style='warn'))

def show_console(cell, rule = False, verbose = False, no_show = False):
    if rule:
        console.print(Rule(cell.type_name() if verbose else ''))

    if not no_show:
        cell.show_console(console)

def show_html(cell, f):
    cell.show_html(f)

katex_preamble = r"""
<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/katex@0.16.3/dist/katex.min.css" integrity="sha384-Juol1FqnotbkyZUT5Z7gUPjQ9gzlwCENvUZTpQBAPxtusdwFLRy382PSDx5UUJ4/" crossorigin="anonymous">
<script defer src="https://cdn.jsdelivr.net/npm/katex@0.16.3/dist/katex.min.js" integrity="sha384-97gW6UIJxnlKemYavrqDHSX3SiygeOwIZhwyOKRfSaf0JWKRVj9hLASHgFTzT+0O" crossorigin="anonymous"></script>
<script defer src="https://cdn.jsdelivr.net/npm/katex@0.16.3/dist/contrib/auto-render.min.js" integrity="sha384-+VBxd3r6XgURycqtZ117nYw44OOcIax56Z4dCRWbxyPt0Koah1uHoK0o4+/RRE05" crossorigin="anonymous"></script>
<script>
    document.addEventListener("DOMContentLoaded", function() {
        renderMathInElement(document.body, {
          // customised options
          // • auto-render specific keys, e.g.:
          delimiters: [
              {left: '$$', right: '$$', display: true},
              {left: '$', right: '$', display: false},
              {left: '\\(', right: '\\)', display: false},
              {left: '\\[', right: '\\]', display: true}
          ],
          // • rendering keys, e.g.:
          throwOnError : false
        });
    });
</script>
<script src="https://cdn.jsdelivr.net/npm/katex@0.16.3/dist/contrib/copy-tex.min.js" integrity="sha384-ww/583aHhxWkz5DEVn6OKtNiIaLi2iBRNZXfJRiY1Ai7tnJ9UXpEsyvOITVpTl4A" crossorigin="anonymous"></script>
"""

def show(fn: "input notebook",
         html: "save HTML to a file" = '',
         katex: "include KaTeX in HTML output" = False,
         external: "external zip archive with binary content" = '',
         debug: "show debugging information" = False):
    """Show the contents of the notebook, without evaluating."""
    with open(fn) as f:
        cells = c.parse(f, external, show_only = True, info=info)
    _show(cells, html, katex, debug)

def _show(cells, html, katex, debug):
    output   = lambda cell: show_console(cell, rule = debug, verbose = debug)

    if html:
        f_html = open(html, 'w')
        output = lambda cell: show_html(cell, f_html)

        f_html.write('<!DOCTYPE html>\n')
        f_html.write('<html>\n')
        f_html.write('<head>\n')
        f_html.write('<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/water.css@2/out/light.css">\n')
        if katex:
            f_html.write(katex_preamble)
        f_html.write('<style>\n')
        f_html.write(c.HtmlFormatter().get_style_defs('.highlight'))
        f_html.write('</style>\n')
        f_html.write('</head>\n')
        f_html.write('<body>\n')


    for i,cell in enumerate(cells):
        if not cell.display(): continue
        output(cell)

    if html:
        f_html.write('</body>\n')
        f_html.write('</html>\n')

@argh.arg('infn', nargs='?')
@argh.arg('outfn', nargs='?')
@argh.arg('--no-mpi')
@argh.arg('-n', '--dry-run')
@argh.arg('-i', '--interactive')
def run(infn: "input notebook",
        outfn: "output notebook (if empty, input modified in place)",
        clean: "run from scratch, ignoring checkpoints" = False,
        auto_capture: "automatically capture images" = False,
        external: "external zip archive with binary content" = '',
        debug: "show debugging information" = False,
        no_mpi: "disable MPI awareness" = False,
        dry_run: "don't save the processed notebook" = False,
        only_root_output: "suppress output everywhere but rank 0 (for MPI)" = False,
        interactive: "run REPL after the notebook is processed" = False):
    """Run the notebook."""

    global using_mpi
    global root

    if not no_mpi:
        try:
            from mpi4py import MPI
            root = MPI.COMM_WORLD.Get_rank() == 0
            using_mpi = MPI.COMM_WORLD.Get_size() > 1
        except:
            pass

    if infn and os.path.exists(infn):
        with open(infn) as f:
            cells = c.parse(f, external, info=info)
    else:
        cells = []
        if outfn and not dry_run:
            warn(f"Input file [error]{infn}[/error] doesn't exist, but given an output file [cyan]{outfn}[/cyan]; forcing [affirm]--dry-run[/affirm]")
            dry_run = True

    if not infn:
        interactive = True

    if not outfn:
        outfn = infn

    sys.argv = [infn] + argv

    def output(cell):
        if root or (not only_root_output and type(cell) is c.OutputCell):
            show_console(cell, rule = debug, verbose = debug)

    nb = notebook.Notebook(name = infn, auto_capture = auto_capture, dry_run = dry_run)
    nb.add(cells)

    if not clean:
        checkpoint = nb.find_checkpoint()
        if checkpoint is not None:
            info(f"Skipping to checkpoint {checkpoint}", style='magenta')
            nb.skip(checkpoint, output)
            info('Resuming', style="magenta")

    try:
        sys.path.insert(0, os.path.dirname(infn))
        nb.process_all(output,
                       repl=lambda: run_repl(nb, output, debug=debug,
                                             prefix = [c.Blanks.create(1)], suffix = [c.Blanks.create(1), c.BreakCell.create()]),
                       force=interactive, info=info, debug=debug)

        if interactive:
            result = run_repl(nb, output, debug=debug)
            if result and not nb.dry_run and not outfn:
                from prompt_toolkit import prompt
                from prompt_toolkit.completion import PathCompleter
                outfn = prompt("Notebook filename (empty to not save): ", completer = PathCompleter())
                if outfn and not external:
                    external = prompt("External zip archive filename (empty to inline): ", completer = PathCompleter())
    except:
        nb.move_all_incoming()

    if not nb.dry_run and root and outfn:
        nb.save(outfn, external)


def run_repl(nb, output, debug = False,
             prefix = [c.Blanks.create(1), c.BreakCell.create(), c.Blanks.create(1)],
             suffix = []):
    if skip_repl:
        return

    if using_mpi:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD

    def execute_line(line):
        line += '\n'
        if using_mpi and root:
            line = comm.bcast(line, root = 0)
        cells = []
        if nb.current > 0:
            cells += prefix
        cells += c.parse(io.StringIO(line), None, info=info)
        cells += suffix

        nb.insert(cells)
        nb.process_to(nb.current + len(cells),output,info=info)

    if not root:
        while True:
            line = comm.bcast('', root = 0)
            if not line: return
            execute_line(line)

    saturn_dir = os.path.expanduser('~/.saturn')
    if not os.path.exists(saturn_dir):
        os.makedirs(saturn_dir)

    try:
        g = {}  # fake empty globals;
                # not specifying any get_globals throws an exception inside repl.run();
                # specifying nb.globals breaks checkpointing inside repl; not sure why
        repl = PythonReplWithExecute(
            execute = execute_line,
            get_globals=lambda: g,
            get_locals=lambda: nb.l,
            vi_mode=False,
            history_filename=saturn_dir + '/history',
            startup_paths=None,
            debug=debug,
        )
    except Exception as e:
        print("Caught exception setting up REPL:", e)
        raise

    # Add the code cells to history
    for cell in nb.cells:
        for line in cell.repl_history():
            repl.history.append_string(line)

    result = True

    @repl.add_key_binding('c-q')
    def _(event):
        nonlocal result
        event.app.exit(exception=EOFError)
        result = False

    # TODO: think about changing this to a toggle
    @repl.add_key_binding('f10')
    def _(event):
        nonlocal result
        nb.dry_run = True
        event.app.exit(exception=EOFError)
        result = False

    @repl.add_key_binding('c-w')
    def _(event):
        global skip_repl
        skip_repl = True
        event.app.exit(exception=EOFError)

    try:
        repl.run()
    except Exception as e:
        print("Caught exception in REPL:", e)
        raise

    if using_mpi:
        comm.bcast('', root = 0)

    return result

@argh.arg('outfn', nargs='?')
def clean(infn: "input notebook",
          outfn: "output notebook (if empty, input modified in place)",
          strip_output: "also strip all output" = False):
    """Remove all binary data from the notebook."""
    if not outfn:
        outfn = infn

    if os.path.exists(infn):
        with atomic_write(outfn, mode='w', overwrite=True) as of:
            with open(infn) as f:
                pf = peekable(f)
                for line in pf:
                    if strip_output and line.startswith('#o>'): continue
                    if line.startswith('#o> png'): continue
                    if line.startswith('#chk>') and line.strip() != '#chk>':
                        of.write('#chk>\n')
                        while pf and pf.peek().startswith('#chk>'):
                            next(pf)
                        continue
                    if line.startswith('#var>'):
                        # Keep the first line, but skip all subsequent lines
                        while pf and pf.peek().startswith('#var>'):
                            next(pf)
                    of.write(line)

@argh.arg('i',   nargs='?', type=int)
@argh.arg('out', nargs='?')
def image(infn: "input notebook", i: "image index", out: "output PNG filename",
          external: "external zip archive with binary content" = '',
          ):
    """Extract an image from the notebook."""
    if i is not None and not out:
        console.print("Must specify output filename, if image is specified")
        return

    with open(infn) as f:
        cells = c.parse(f, external, show_only = True, info=info)

    count = 0
    for cell in cells:
        if type(cell) is not c.OutputCell: continue

        for x in cell.composite_:
            if type(x) is io.StringIO: continue

            if i is None:
                print(f"{count}:")
                show_png(x)
            elif i == count:
                with open(out, 'wb') as f:
                    f.write(x)

            count += 1

def version():
    """Show version of Saturn and its dependencies."""
    from importlib_metadata import version as ver
    print(f"Saturn {ver('saturn_notebook')} (Python {sys.version})")
    for dep in ['wurlitzer', 'rich', 'ptpython',
                'dill', 'markdown', 'atomicwrites',
                'pygments', 'more_itertools', 'matplotlib', 'nbformat']:
        print(f"   {dep} {ver(dep)}")

@argh.arg('outfn', nargs='?')
def convert(infn: "Jupyter notebook",
            outfn: "output notebook (if empty, show the cells instead)",
            version: "notebook version" = 4,
            external: "external zip archive with binary content" = '',
            html: "save HTML to a file" = '',
            katex: "include KaTeX in HTML output" = False,
            debug: "show debugging information" = False):
    """Convert a Jupyter notebook into a Saturn notebook."""
    import nbformat
    jnb = nbformat.read(infn, as_version=version)

    header = c.CodeCell()
    header.lines_ = ['# vim: ft=python foldmethod=marker foldlevel=0\n']
    cells = [header]

    for jcell in jnb.cells:
        if jcell['cell_type'] == 'markdown':
            cell = c.MarkdownCell()
            cell.lines_ = [' ' + line + '\n' if len(line) else '\n' for line in jcell['source'].split('\n')]
            if type(cells[-1]) is not c.Blanks:
                cells.append(c.Blanks.create(1))
            cells.append(cell)
            cells.append(c.Blanks.create(1))
        elif jcell['cell_type'] == 'code':
            if type(cells[-1]) is c.CodeCell:
                cells.append(c.Blanks.create(1))
                cells.append(c.BreakCell.create())
                cells.append(c.Blanks.create(1))
            cell = c.CodeCell()
            cell.lines_ = [line + '\n' for line in jcell['source'].split('\n')]
            cells.append(cell)

            for out in jcell['outputs']:
                if out['output_type'] == 'stream':
                    cell = c.OutputCell.from_string(out['text'])
                    cells.append(cell)
                elif out['output_type'] in ['display_data', 'execute_result']:
                    if 'image/png' in out['data']:
                        cell = c.OutputCell()
                        png_content = out['data']['image/png']
                        cell.composite_.append_png(c.base64.b64decode(png_content))
                        cells.append(cell)
                    elif 'text/plain' in out['data']:
                        cell = c.OutputCell.from_string(out['data']['text/plain'])
                        cells.append(cell)
                    else:
                        info('Unrecognized data type', style="magenta")
                else:
                    info('Unrecognized output type', style="magenta")
        else:
            info('Unrecognized cell type', style="magenta")

    if not outfn:
        _show(cells, html, katex, debug)
    else:
        nb = notebook.Notebook(name = outfn)
        nb.add(cells)
        nb.move_all_incoming()
        nb.save(outfn, external)


@argh.arg('outfn', nargs='?')
def rehash(infn: "input notebook",
           outfn: "output notebook (if empty, input modified in place)",
           external: "external zip archive with binary content" = ''):
    """Rehash all the code cells, updating the hashes stored with checkpoints and variable cells. (advanced)"""
    if not outfn:
        outfn = infn

    with open(infn) as f:
        cells = c.parse(f, external, info=info)

    nb = notebook.Notebook(name = infn)
    nb.add(cells)
    nb.rehash()

    nb.save(outfn, external)

@argh.arg('outfn', nargs='?')
def extract(infn: "input notebook",
            external: "external zip archive with binary content",
            outfn: "output notebook (if empty, input modified in place)"):
    """Extract embedded binary content into external zip archive."""
    if not outfn:
        outfn = infn

    with open(infn) as f:
        cells = c.parse(f, '', info=info)       # read without external

    nb = notebook.Notebook(name = infn)
    nb.add(cells)
    nb.move_all_incoming()

    nb.save(outfn, external)

@argh.arg('outfn', nargs='?')
def embed(infn: "input notebook",
          external: "external zip archive with binary content",
          outfn: "output notebook (if empty, input modified in place)"):
    """Embed binary content from external zip archive into notebook body."""
    if not outfn:
        outfn = infn

    with open(infn) as f:
        cells = c.parse(f, external, info=info)

    nb = notebook.Notebook(name = infn)
    nb.add(cells)
    nb.move_all_incoming()

    nb.save(outfn, '')      # write without external

@argh.arg('--no-mpi')
@argh.arg('-n', '--dry-run')
def _run(clean: "run from scratch, ignoring checkpoints" = False,
        auto_capture: "automatically capture images" = False,
        external: "external zip archive with binary content" = '',
        debug: "show debugging information" = False,
        no_mpi: "disable MPI awareness" = False,
        dry_run: "don't save the processed notebook" = False,
        only_root_output: "suppress output everywhere but rank 0 (for MPI)" = False):
    """Launch Saturn REPL."""
    run('', '', clean, auto_capture, external, debug, no_mpi, dry_run, only_root_output, True)

def main():
    global argv

    parser = argh.ArghParser()
    parser.set_default_command(_run)

    if '--' in sys.argv:
        idx = sys.argv.index('--')
        argv = sys.argv[idx+1:]
        sys.argv = sys.argv[:idx]
    parser.add_commands([show, run, clean, image, version, convert, rehash, extract, embed])
    parser.dispatch()

if __name__ == '__main__':
    main()



from    rich.theme   import Theme
from    rich.style   import Style

theme = Theme({
    "variables":    Style.parse("yellow"),
    "warn":         Style.parse("yellow"),
    "affirm":       Style.parse("green"),
    "error":        Style.parse("red"),
    "cell-name":    Style.parse("yellow"),
})



from rich.markdown import Markdown
from rich.syntax   import Syntax
from rich.text     import Text
from rich.rule     import Rule
from rich.abc      import RichRenderable

from  itertools import chain
import base64
import io
import dill
import hashlib
from  more_itertools    import peekable

import html
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import HtmlFormatter

import markdown

from . import utils
from . import image
from . import evaluate

import  zipfile, os

# Prefix for the filename inside an external zipfile
_zip_fn_prefix = 'name='

def hash_bytes(content):
    return hashlib.sha256(content).hexdigest()[:16]

class Cell:
    def __init__(self):
        self.lines_ = []

    def append(self, line):
        prefix = self.__class__._prefix
        line = line[len(prefix):]       # eat the prefix
        self.lines_.append(line)

    def save(self, external):
        prefix = self.__class__._prefix
        return [prefix + line for line in self.lines_]

    def parse(self, external, info):            # called after all the lines have been read
        pass

    def lines(self):
        return ''.join(self.lines_)

    def repl_history(self):
        return []

    @classmethod
    def display(cls):
        return True

    def empty(self):
        return not self.__class__.display()

    @classmethod
    def identify(cls, line):
        return line.startswith(cls._prefix)

    def type_name(self):
        if hasattr(self.__class__, '_name'):
            return self.__class__._name
        else:
            return self.__class__.__name__

    def __repr__(self):
        return '== ' + self.__class__.__name__ + '\n' + self.lines()

    def __rich__(self):
        return '[cell-name]== ' + self.__class__.__name__ + ' ==[/cell-name]\n' + self.lines()

    def show_console(self, console):
        console.print(self)

    def show_html(self, f):
        f.write(self._render_html())
        f.write('\n')

    def _render_html(self):
        return f"<div><h2>{self.__class__.__name__}</h2> <pre>{html.escape(self.lines())}</pre></div>"

class CodeCell(Cell):
    _prefix = ''
    _name   = 'Code'

    def __init__(self):
        super().__init__()
        self.skippable = True
        self.hashable  = True

    def __rich__(self):
        if not self.lines_:
            return
        return Syntax('>>> ' + '... '.join(self.lines_).strip(), 'python')

    def append(self, line):
        super().append(line)
        if line.startswith('#no-skip#'):
            self.skippable = False
        if line.startswith('#no-hash#'):
            self.hashable = False

    def code(self):
        lines = self.lines()
        if len(lines) > 0 and lines[-1] == '\n':
            return lines[:-1]
        return lines

    def _render_html(self):
        return f"<div class='code'>{highlight(self.code(), PythonLexer(), HtmlFormatter())}</div>"

    def repl_history(self):
        return self.save(None)

class MarkdownCell(Cell):
    _prefix = '#m>'
    _name   = 'Markdown'

    def __rich__(self):
        return Markdown(self.lines())

    def _render_html(self):
        return "<div class='markdown'>" + markdown.markdown(''.join(line[1:] if line[0] == ' ' else line for line in self.lines_)) + "</div>"

class OutputCell(Cell):
    _prefix = '#o> '
    _name   = 'Output'

    def __init__(self, composite_ = None):
        super().__init__()
        if composite_ is None:
            self.composite_ = utils.CompositeIO()
        else:
            self.composite_ = composite_

    @staticmethod
    def from_string(s):
        cell = OutputCell()
        cell.composite_.write(s)
        return cell

    def parse(self, external, info):
        super().parse(external, info)
        pl = peekable(self.lines_)
        for line in pl:
            if not line.startswith('png'):
                self.composite_.write(line)
            else:
                if _zip_fn_prefix in line:
                    # take everything from name= to the end (-1 to not include \n)
                    fn = line[line.index(_zip_fn_prefix) + len(_zip_fn_prefix):-1]
                    if external:
                        if zipfile.Path(external, at=fn).exists():
                            png_content = external.read(fn)
                            self.composite_.append_png(png_content)
                        else:
                            # here and below, we use
                            # self.composite_.append_rich, instead of info, so
                            # that the message appears in place of the cell,
                            # rather than before any other cell is processed
                            self.composite_.append_rich(Rule(f"[warn]image hash [cyan]{fn}[/cyan] not found in the external archive [cyan]{external.filename}[/cyan][/warn]"))
                    else:
                        self.composite_.append_rich(Rule(f"[warn]image hash found, but no external file given ([cyan]{fn}[/cyan])[/warn]"))
                else:
                    png_content = [line[3:]]
                    while pl and pl.peek().startswith('png') and not pl.peek().startswith('png{{{'):
                        png_content.append(next(pl)[3:])
                    png_content = ''.join(png_content)
                    self.composite_.append_png(base64.b64decode(png_content))
        self.lines_.clear()

    def save(self, external):
        lines_ = []
        for x in self.composite_:
            if type(x) is io.StringIO:
                x.seek(0)
                lines_ += [self._prefix + line for line in utils.collapse_carriage_return(x)]
            else:
                if external:
                    fn = f"{hash_bytes(x)}.png"
                    with external.open(fn, 'w') as f:
                        f.write(x)
                    lines_ += [self._prefix + f'png {_zip_fn_prefix}{fn}\n']
                else:   # inline
                    content = base64.b64encode(x).decode('ascii')
                    lines_ += [self._prefix + 'png' + line + '\n' for line in chunk(content, 80, markers = True)]
        return lines_

    def empty(self):
        return self.composite_.empty()

    def show_console(self, console):
        for x in self.composite_:
            if type(x) is io.StringIO:
                if x.getvalue():
                    console.print(Text(x.getvalue()), end='')
            elif type(x) is bytes:
                image.show_png(x)
            elif isinstance(x,RichRenderable):
                console.print(x)
            else:
                console.print(Rule("[error]didn't recognize cell type[/error]"))

    def _render_html(self):
        result = ""
        for x in self.composite_:
            if type(x) is io.StringIO:
                result += f"<div class='output'><pre>{html.escape(x.getvalue())}</pre>\n"
            else:
                result += f'<img src="data:image/png;base64,{base64.b64encode(x).decode("ascii")}"/>\n'
        result += "</div>"
        return result

class BreakCell(Cell):
    _prefix = '#---#'

    @classmethod
    def display(cls):
        return False

    @staticmethod
    def create():
        cell = BreakCell()
        cell.lines_ = ['\n']
        return cell


class REPLCell(Cell):
    _prefix = '#-REPL-#'

    @classmethod
    def display(cls):
        return False

class SaturnCell(Cell):
    _prefix = '#saturn>'

    _external_prefix = 'external='

    def parse(self, external, info):            # called after all the lines have been read
        super().parse(external, info)
        line = self.lines_[0]
        self.external_fn = ''
        if self._external_prefix in line:
            # take everything from external= to the end (-1 to not include \n)
            # eventually, we'll want a more elaborate support, once we have other kinds of metadata
            self.external_fn = line[line.index(self._external_prefix) + len(self._external_prefix):-1]

    @classmethod
    def display(cls):
        return False

class CheckpointCell(Cell):
    _prefix = '#chk>'
    _extension = '.chk'
    _warning_name = "checkpoint"

    def expected(self, h):
        if self.expected_hash() == None:
            return False
        else:
            return self.expected_hash() == h

    def expected_hash(self):
        if not hasattr(self, '_expected'):
            self._expected = dill.load(self._content)
        return self._expected

    def show_console(self, console):
        if hasattr(self, '_warning'):
            console.print(self._warning)

    def parse(self, external, info):
        if all(l.strip() == '' for l in self.lines_):
            self.lines_ = []

        if not self.lines_:
            self._expected = None
            return

        line = self.lines_[0]
        if _zip_fn_prefix in line:
            fn = line[line.index(_zip_fn_prefix) + len(_zip_fn_prefix):-1]
            if external:
                if zipfile.Path(external, at=fn).exists():
                    self._content = io.BytesIO(external.read(fn))
                else:
                    self._warning = Rule(f"[warn]{self._warning_name} hash [cyan]{fn}[/cyan] not found in the external archive [cyan]{external.filename}[/cyan][/warn]")
                    self._expected = None
            else:
                self._warning = Rule(f"[warn]{self._warning_name} hash found, but no external file given ([cyan]{fn}[/cyan])[/warn]")
                self._expected = None
        else:
            content = self.lines()
            content = base64.b64decode(content)
            self._content = io.BytesIO(content)

    def load(self):
        h = self.expected_hash()    # just to make sure it's been parsed
        assert h != None
        if not hasattr(self, '_locals'):
            self._locals = dill.load(self._content)
        return self._locals

    def dump(self, running, locals_):
        try:
            content = io.BytesIO()
            dill.dump(running, content)
            dill.dump(locals_, content)
            self._content = content
        except:
            self._content = None
            raise

    def save(self, external):
        if hasattr(self, '_content') and self._content:
            content = self._content.getvalue()
            if not external:
                content = base64.b64encode(content).decode('ascii')
                self.lines_ = [line + '\n' for line in chunk(content, 80, markers = True)]
            else:
                fn = f"{hash_bytes(content)}{self._extension}"
                with external.open(fn, 'w') as f:
                    f.write(content)
                self.lines_ = [f' {_zip_fn_prefix}{fn}\n']
        else:
            self.lines_ = ['']      # to keep the blank checkpoint cell
        return super().save(external)

    def replace_hash(self, running):
        if self.expected_hash() == None: return
        content = io.BytesIO()
        dill.dump(running, content)
        content.write(self._content.read())
        self._content = content

class VariableCell(CheckpointCell):
    _prefix = '#var>'
    _extension = '.var'
    _warning_name = "variable"

    def parse(self, external, info):
        self.variables = self.lines_[0]
        self.lines_ = self.lines_[1:]

        super().parse(external, info)

    def load(self, locals_):
        locals_['_var_cell_load'] = dill.load(self._content)
        evaluate.exec_eval(self.variables.strip() + ' = _var_cell_load', locals_, locals_)
        del locals_['_var_cell_load']

    def dump(self, running, locals_):
        content = io.BytesIO()
        dill.dump(running, content)
        dill.dump(evaluate.eval_expression(self.variables, locals_), content)
        self._content = content

    def header(self):
        return self.__class__._prefix + self.variables

    def save(self, external):
        lines_ = super().save(external)
        return [self.header()] + lines_

    def repl_history(self):
        return [self.header()]

# Captures blank lines between cells
class Blanks(Cell):
    _prefix = ''

    def show_console(self, console):
        for line in self.lines_:
            console.print(line, end='')

    def show_html(self, f):
        pass

    @staticmethod
    def create(n):
        cell = Blanks()
        cell.lines_ = ['\n']*n
        return cell

cell_types = [MarkdownCell, OutputCell, BreakCell, CheckpointCell, VariableCell, REPLCell, SaturnCell]

def identify(line):
    for Type in chain(cell_types, [CodeCell]):        # CodeCell matches everything, so comes last
        if Type.identify(line):
            return Type

def chunk(content, width, markers = False):
    chunking = utils.chunkstring(content, 80)
    if markers:
        chunking = chain(['{{{'], chunking, ['}}}'])
    return chunking

def open_external(external_fn, show_only, info):
    if external_fn:
        if not os.path.exists(external_fn):
            if show_only:
                info(f"External zip archive [error]{external_fn}[/error] not found.", style='warn')
            return None

        return zipfile.ZipFile(external_fn, 'r')
    else:
        return None

def parse(f, external_fn, *, show_only = False, info = lambda *args, **kwargs: None):
    external = open_external(external_fn, show_only, info)

    def should_parse(cell):
        # skip check-point cells in show mode
        return not show_only or cell.display() or type(cell) is SaturnCell

    cells = []
    def cells_append(cell):
        nonlocal external, external_fn
        if len(cells) > 0 and should_parse(cells[-1]):
            cells[-1].parse(external, info)

            if type(cells[-1]) is SaturnCell:
                if not external_fn and cells[-1].external_fn:
                    external_fn = cells[-1].external_fn
                    external = open_external(external_fn, show_only, info)

        cells.append(cell)

    p = peekable(f)
    for line in p:
        # agglomerate empty lines into Blanks and either store as such or return to the CodeCell, if in the middle of one
        if not line.strip():
            blank = Blanks()
            blank.append(line)
            while p and not p.peek().strip():
                blank.append(next(p))

            if p and len(cells) > 0 and type(cells[-1]) is CodeCell and identify(p.peek()) is CodeCell:
                for line in blank.lines_:
                    cells[-1].append(line)
            else:
                cells_append(blank)

            continue

        Type = identify(line)

        if len(cells) == 0 or type(cells[-1]) is not Type:
            cells_append(Type())
        cells[-1].append(line)

    if len(cells) > 0:
        if should_parse(cells[-1]):
            cells[-1].parse(external, info)

    return cells
